	Tracing temporal expressions
	============================

This tracer works by enetring automatically tracing printouts using the 
'exec' construct.
The printout points will be
  - after a success of a sequence step, (before the ';' and '}')
  - after a success of the lhs of a yield (before the '=>')
  - after a success of 'or' operands
  - conditionally after success of a repeated part


Usage:

1. Before loading the temporals to be traced :

   - Load this file. This will automatically add "tracing execs"
     to any temporals loaded from now on.

   - If you also want tracing of repeat parts (e.g. the "@a" in "[8] * @a")
     then type the following before loading the temporals:

	trace te -repeat


2. Load your files

3. To turn on tracing for specific temporals type:

 	tr[ace] te [[on] [line]<line> [in] @<module>] [if (<bool-exp>)]


   This trace can be deleted using "delete break", and shown using
   "show break".

   The optional <bool-exp> can use 'the_struct' to refer to the struct 
   instance where the TE is.


o Examples:

  trace te on 30 in @checker 
  trace te on 30 in @checker if (the_struct == packet-@58 and sys.time > 100) 
  trace te on 17 in @checker if (the_struct.as_a(packet).id != 0)

o An example of a trace printout:

-> 9 te packet-@0 ([41] * @bbb) or ([43] * @bbb) #3
          event zz is { @aa;                             at line 32 in @chkr

-> <time> te <instance>: <the sub_te which currently succeeded> #<trace-point>
	  <first line of the te>			 <source reference>

  The #trace_point is another indication which sub_te currently succeeded
  for the cases in which the same sub_te appears more than once in the te,
  as in {@aa;@bb;@aa}.

  In this case, the two occurances of "@aa" will have different trace
  points, namely #1 and #3:
  
	{(@aa) #1;(@bb) #2;(@aa) #3}

  To see the te with trace points details type:

 	sh[ow] te [on] [line] <line> [in] @<module>

Notes:
  When running on top of version 4.0, the tracer may cause some complex
  temporals to exceed the capacity limit. 
  In such cases that temporal should be commented out to trace other
  temporals.

//-------------------------------------------------------------------------

<'
define <short_trace_te'command> 
    "tr[ace] te[ on][ line] <num>[ in] [@]<wild_name>[ if <exp> ]" as {
         trace te if (module == "<wild_name>" and line == <num> and <exp|TRUE>);
};

define <show_trace_te'command> 
    "sh[ow] te[ on][ line] <num>[ in] [@]<wild_name>" as {
    scheduler.show_trace_te("<wild_name>",<num>)
};

define <trace_te'command> 
    "tr[ace] te[ if <exp>]" as {
    trace scheduler.trace_te if 
	    (scheduler.ok_te(<exp|TRUE>));
};

define <trace_te_enable'command> "tr[ace] te<opt>(repeat)" as {
     scheduler.debug_te_repeat = ("<3>" == "repeat");
};


// ******************************
// trace_te_str
//
// Holds the string with the points
// for a specific te
// ******************************
struct trace_te_str {
    line    	: uint;
    module	: module;
    str    	: string;
}; // trace_te_str


// ******************************
// scheduler
// ******************************
extend scheduler {
    debug_tes: bool;  		// If true, we create the extra code
    debug_te_repeat: bool;	// If true, we trace repeated parts
    debug_te_sr: source_item;
    debug_te_next: bool;
    debug_te_point:int;
    debug_te_source_line:int;
    debug_te_source_module:module;
    trace_te_strs : list of trace_te_str;

    //----------------------------------------------------------
    // show_trace_te()
    //
    // shows the te giving in detail the trace points 
    //----------------------------------------------------------
    show_trace_te(module_wild_name : string,line : int) is {
	out("line ", line, " in @", module_wild_name );
        var trace_te_id : int = trace_te_strs.last_index(
            (it.line == line) and
            (it.module.name ~ module_wild_name) );
        if trace_te_id != UNDEF {
	    var chopped_str := str_split(trace_te_strs[trace_te_id].str, ";");
	    for each (s) in chopped_str {
		var ns := str_replace( s, "/\(\((.`)\) #(.`)\)/", "(\\1) #\\2" );
		out(ns, ";" );
	    };
        } else {
            error("TE not found.");
        };
    }; // show_trace_te


    //----------------------------------------------------------
    // record_trace_te()
    //
    // Records the te after entering the trace points,  in favor 
    // of later showing it giving in detail the trace points 
    //----------------------------------------------------------
    record_trace_te( str : string, current_ntv : ntv ) is {
        var trace_te_str : trace_te_str = trace_te_strs.last(
            (it.line == debug_te_source_line) and
            (it.module == debug_te_source_module) );
	var clean_str := str_replace( str, 
		"/exec {scheduler.trace_te\((.`),(.`),(.`),.`}/", "#\\1" );
	var full_clean_str := str_expand_dots( clean_str );
        if trace_te_str == NULL {
	    trace_te_str = new trace_te_str (tts) with {
		tts.module = debug_te_source_module;
		tts.line = debug_te_source_line;
	    };
	    trace_te_strs.add( trace_te_str );
	    trace_te_str.str = full_clean_str;
	} else {
	    var full_orig_str := str_expand_dots( current_ntv.x.conv_orig_string );
	    var top_split := str_split_all( trace_te_str.str, full_orig_str ); 
	    var to_replace_idx := top_split.first_index( it == full_orig_str );
	    if to_replace_idx != UNDEF {
	        top_split[ to_replace_idx ] = full_clean_str;
	        trace_te_str.str = str_join( top_split, "");
	//    } else {
	//        out( "trace_te_str.str ", trace_te_str.str );
	//	out( "full_orig_str ", full_orig_str );
	    };
	};
	// out( "R ", trace_te_str.line, " @", trace_te_str.module.name, " ", trace_te_str.str );
    }; // record_trace_te

    
    //----------------------------------------------------------
    // ok_te()
    //
    // Checks dynamically the condition the user gave, to
    // determine whether to output the te trace printouts
    //----------------------------------------------------------
    ok_te(b: bool): bool is {
	if b {
	    debug_te_next = TRUE;
	};
        return FALSE;
    }; // ok_te


    //----------------------------------------------------------
    // trace_te()
    //
    // Outputs the te trace printouts 
    //----------------------------------------------------------
    trace_te(point: int, te: string, line: int, module: string, 
					the_struct: any_struct) is {
	if not debug_te_next {return};
	debug_te_next = FALSE;
	debug_te_sr.set_source_ref(symtab.modules.key(module), line);
	out("-> ", sys.time, " te ", the_struct, ":  ", te, " #", dec(point));
	symtab.show_source(debug_te_sr);
    }; // trace_te


    //----------------------------------------------------------
    // trace_te_define_init()
    //
    // Returns whether current ntv string should be reparsed
    // and if we moved to a new te inits its point to start from 1
    //----------------------------------------------------------
    trace_te_define_init( s:string, line:int, module:module ):bool is {
        if not scheduler.debug_tes or s ~ "/trace_te/" then {
	    return FALSE;
    	};
	result = TRUE;
        if line != debug_te_source_line or module != debug_te_source_module {
            debug_te_point = 1;
            debug_te_source_line = line;
            debug_te_source_module = module;
        };
    }; // trace_te_define_init


    //----------------------------------------------------------
    // init()
    //----------------------------------------------------------
    init() is also {
        debug_te_sr = new;
        debug_tes = TRUE;
    }; // init

}; // scheduler


// define <for_debug_repeat'tte> "[~][ ]\[<exp>\.\.<exp>\][ ]*[ ]<tte>" as computed {
// ******************************
// 	repeat 
// ******************************
define <for_debug_repeat'tte> "[~][ ]\[<exp>\][ ][\*][ ]<tte>" as computed {
    var full := str_expand_dots(x.string);
    var line := me.source_line();
    var module := me.source_module();
    if not scheduler.trace_te_define_init( full, line, module ) then {
	error_backtrack("");  -- Ignore this define
    };
    if not scheduler.debug_te_repeat then {
        scheduler.record_trace_te( me.x.string, me );
	error_backtrack("");  -- Ignore this define
    };
    var split_by_repeat := str_split( x.string, "/\*/" );
    var repeat : string;
    var repeated_part : string;
    if split_by_repeat.size() == 1 {
	full = append( full, " * cycle" );
	repeated_part = "cycle";
	repeat = full;
    } else {
        repeat = split_by_repeat[0];
        repeated_part = split_by_repeat[1];
    };
    result = append(repeat, " * ", 
		    "((", repeated_part, ") exec {scheduler.trace_te(",
		    dec(scheduler.debug_te_point), ",", quote(repeated_part), 
		    ", ", line, ", ", quote(module.name),", me)})");
    scheduler.record_trace_te( result, me );
    scheduler.debug_te_point += 1;
}; // for_debug_repeat


// ******************************
// 	yield  =>
// ******************************
define <for_debug_yield'tte> "<t1'tte>[ ]=\>[ ]<t2'tte:long>" as computed {

    var full1 := str_expand_dots(<1>);
    var line := me.source_line();
    var module := me.source_module();
    if not scheduler.trace_te_define_init( full1, line, module ) then {
	error_backtrack("");  -- Ignore this define
    };
    var full2 := str_expand_dots(<4>);
    result = append("((", full1, ") exec {scheduler.trace_te(",
		    dec(scheduler.debug_te_point), ",", 
		    quote(full1), ", ", me.source_line(), ", ",
		    quote(me.source_module().name),	", me)})",
    		    " => ", full2 );
    scheduler.record_trace_te( result, me );
    scheduler.debug_te_point += 1;
}; // for_debug_yield

 
// ******************************
// 	or 
// ******************************
define <for_debug_or'tte> "<t1'tte:long> or <t2'tte>" as computed {

    var full1 := str_expand_dots(<1>);
    var line := me.source_line();
    var module := me.source_module();
    if not scheduler.trace_te_define_init( full1, line, module ) then {
	error_backtrack("");  -- Ignore this define
    };
    var full2 := str_expand_dots(<2>);
    result = append("((", full1, ") exec {scheduler.trace_te(",
		    dec(scheduler.debug_te_point), ",", 
		    quote(full1), ", ", me.source_line(), ", ",
		    quote(me.source_module().name),	", me)})",
    		    " or ((", full2, ") exec {scheduler.trace_te(",
		    dec(scheduler.debug_te_point+1), ",", 
		    quote(full2), ", ", me.source_line(), ", ",
		    quote(me.source_module().name),	", me)})"
		   );
    scheduler.debug_te_point += 2;
    scheduler.record_trace_te( result, me );
    // out(" OR result ", result )
}; // for_debug_or


// ******************************
// 	sequence 
// ******************************
define <for_debug_seq'tte> "({<any>})" as computed {

    var full:= str_expand_top_level_dots(<1>,FALSE);
    var line := me.source_line();
    var module := me.source_module();
    if not scheduler.trace_te_define_init( str_expand_dots(<1>), line, module ) {
	error_backtrack("");  -- Ignore this define
    };
    full = str_replace(full, "/{(.*)}/", "\\1");
    // split by chop
    var parts : list of string;
    parts = str_split_all(full, "/;/");

    -- Now replace
    for each (p) in parts do {
        full = str_expand_dots(p);
	if full != "" and full != ";" and
		(full !~ "/\\[(.`)\.\.(.`)\\] ?(\*)*/" or   // not tm|fm
	         full ~ "/(~\\[(.*)\.\.(.*)\\] ?(\*)*)/" or // tm
	         full ~ "/;/") {                            // in sub sequence
            parts[index] = append("((",full,
		    ") exec {scheduler.trace_te(",
		    dec(scheduler.debug_te_point), ",", 
		    quote(full), ", ", me.source_line(), ", ",
		    quote(me.source_module().name),	", me)})" );
            scheduler.debug_te_point += 1;
	};
    };
    result = append("{", str_join(parts, ""), "}"); 
    scheduler.record_trace_te( result, me );
    // out(" SEQ result\n ", result );
}; // for_debug_seq

'>


